@@ -1,15 +1,12 @@
 # SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
-# SPDX-FileCopyrightText: 2022 Anurag Bansal <anurag.bansal.585@gmail.com>
 # SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause
 
 import math
 import ctypes
+import weakref
+import logging
 from ctypes import c_float
 import pypdfium2._pypdfium as pdfium
-from pypdfium2._helpers._utils import (
-    validate_colours,
-    auto_bitmap_format,
-)
 from pypdfium2._helpers.misc import (
     OptimiseMode,
     PdfiumError,
@@ -20,6 +17,9 @@ from pypdfium2._helpers.misc import (
     BitmapTypeToStr,
     BitmapTypeToStrReverse,
 )
+from pypdfium2._helpers.pageobject import (
+    PdfPageObject,
+)
 from pypdfium2._helpers.converters import (
     BitmapConvBase,
     BitmapConvAliases,
@@ -31,6 +31,8 @@ try:
 except ImportError:
     harfbuzz = None
 
+logger = logging.getLogger(__name__)
+
 
 class PdfPage (BitmapConvAliases):
     """
@@ -44,13 +46,34 @@ class PdfPage (BitmapConvAliases):
     def __init__(self, raw, pdf):
         self.raw = raw
         self.pdf = pdf
+        # if the form env of the parent document is initialised, we could call FORM_OnAfterLoadPage() here
+        self._finalizer = weakref.finalize(
+            self, self._static_close,
+            self.raw, self.pdf,
+        )
+    
+    def _tree_closed(self):
+        if self.raw is None:
+            return True
+        return self.pdf._tree_closed()
+    
+    @staticmethod
+    def _static_close(raw, parent):
+        # logger.debug("Closing page")
+        if parent._tree_closed():
+            logger.critical("Document closed before page (this is illegal). Document: %s" % parent)
+        # if the form env of the parent document is initialised, we could call FORM_OnBeforeClosePage() here
+        pdfium.FPDF_ClosePage(raw)
     
     def close(self):
         """
-        Close the page to release allocated memory.
-        This function shall be called when finished working with the object.
+        Free memory by applying the finalizer for the underlying PDFium page.
+        Please refer to the generic note on ``close()`` methods for details.
         """
-        pdfium.FPDF_ClosePage(self.raw)
+        if self.raw is None:
+            logger.warning("Duplicate close call suppressed on page %s" % self)
+            return
+        self._finalizer()
         self.raw = None
     
     
@@ -89,8 +112,9 @@ class PdfPage (BitmapConvAliases):
     
     def _get_box(self, box_func, fallback_func):
         left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
-        ret_code = box_func(self.raw, left, bottom, right, top)
-        if not ret_code:
+        success = box_func(self.raw, left, bottom, right, top)
+        if not success:
+            # TODO avoid repeated initialisation of c_float objects for fallback
             return fallback_func()
         return (left.value, bottom.value, right.value, top.value)
     
@@ -170,6 +194,42 @@ class PdfPage (BitmapConvAliases):
         return PdfTextPage(textpage, self)
     
     
+    def insert_object(self, pageobj):
+        """
+        Insert a page object into the page.
+        
+        Position and form are defined by the object's matrix.
+        If it is the identity matrix, the object will appear as-is on the bottom left corner of the page.
+        
+        The page object must not belong to a page yet. If it belongs to a PDF, this page must be part of the PDF.
+        
+        You may want to call :meth:`.generate_content` once you finished adding new content to the page.
+        
+        Parameters:
+            pageobj (PdfPageObject): The page object to insert.
+        """
+        
+        if pageobj.page is not None:
+            raise ValueError("The pageobject you attempted to insert already belongs to a page.")
+        if (pageobj.pdf is not None) and (pageobj.pdf is not self.pdf):
+            raise ValueError("The pageobject you attempted to insert belongs to a different PDF.")
+        
+        pdfium.FPDFPage_InsertObject(self.raw, pageobj.raw)
+        
+        pageobj.page = self
+        pageobj.pdf = self.pdf
+    
+    
+    def generate_content(self):
+        """
+        Generate added page content.
+        This function shall be called to apply changes before saving the document or reloading the page.
+        """
+        success = pdfium.FPDFPage_GenerateContent(self.raw)
+        if not success:
+            raise PdfiumError("Generating page content failed.")
+    
+    
     def insert_text(
             self,
             text,
@@ -185,6 +245,8 @@ class PdfPage (BitmapConvAliases):
         Insert text into the page at a specified position, using the writing system's ligature.
         This function supports Asian scripts such as Hindi.
         
+        You may want to call :meth:`.generate_content` once you finished adding new content to the page.
+        
         Parameters:
             text (str):
                 The message to insert.
@@ -200,6 +262,9 @@ class PdfPage (BitmapConvAliases):
                 PDF font data.
         """
         
+        # User-contributed code
+        # SPDX-FileCopyrightText: 2022 Anurag Bansal <anurag.bansal.585@gmail.com>
+        
         hb_buffer = harfbuzz.Buffer()
         hb_buffer.add_str(text)
         hb_buffer.guess_segment_properties()
@@ -218,8 +283,6 @@ class PdfPage (BitmapConvAliases):
             )
             pdfium.FPDFPage_InsertObject(self.raw, pdf_textobj)
             start_point += (pos.x_advance / hb_font.scale) * font_size
-        
-        pdfium.FPDFPage_GenerateContent(self.raw)
     
     
     def get_objects(self, max_depth=2, form=None, level=0):
@@ -255,14 +318,16 @@ class PdfPage (BitmapConvAliases):
             if raw_obj is None:
                 raise PdfiumError("Failed to get page object.")
             
-            helper_obj = PdfPageObject(
+            type = pdfium.FPDFPageObj_GetType(raw_obj)
+            yield PdfPageObject(
                 raw = raw_obj,
+                type = type,
                 page = self,
+                pdf = self.pdf,
                 level = level,
             )
-            yield helper_obj
             
-            if level < max_depth-1 and helper_obj.type == pdfium.FPDF_PAGEOBJ_FORM:
+            if level < max_depth-1 and type == pdfium.FPDF_PAGEOBJ_FORM:
                 yield from self.get_objects(
                     max_depth = max_depth,
                     form = raw_obj,
@@ -275,14 +340,8 @@ class PdfPage (BitmapConvAliases):
         Rasterise a page to a specific output format.
         
         Parameters:
-            
             converter (BitmapConvBase | typing.Callable):
-                
                 A translator to convert the output of :meth:`.render_base`. See :class:`.BitmapConv` for a set of built-in converters.
-                It may be a class, or an instance of a class, that implements the converter interface by inheriting from :class:`.BitmapConvBase` and overriding :meth:`~.BitmapConvBase.run`. Converters may be initialised with parameters to be passed to their :meth:`~.BitmapConvBase.run` method.
-                
-                Alternatively, if your custom converter does not take any parameters, you may also just pass in a function (or callable) that consumes two positional arguments (rendering output, dictionary of rendering keywords) and returns the converted result.
-            
             renderer_kws (dict):
                 Keyword arguments to the renderer.
         
@@ -308,14 +367,10 @@ class PdfPage (BitmapConvAliases):
                 data, cl_format, size = render_to(BitmapConv.any(bytes), ...)
         """
         
-        # In the future, we could add means to set different defaults for specific built-in converters, if necessary.
-        # We could also consider implementing a parameter sieve to automatically divide keyword arguments between converter and renderer so that callers don't need to care about the separation
-        
         args = (self.render_base(**renderer_kws), renderer_kws)
         if isinstance(converter, BitmapConvBase):
             return converter.run(*args, *converter.args, **converter.kwargs)
         elif isinstance(converter, type) and issubclass(converter, BitmapConvBase):
-            # run() is supposed to be a static method, but just initialise an instance of the converter class so it also works if the implementer forgot the decorator
             return converter().run(*args)
         elif callable(converter):
             return converter(*args)
@@ -352,8 +407,8 @@ class PdfPage (BitmapConvAliases):
         Parameters:
             
             scale (float):
-                This parameter defines the resolution of the image.
-                Technically speaking, it is a factor scaling the number of pixels that represent the length of 1 PDF canvas unit (equivalent to 1/72 of an inch by default). [1]_
+                A factor scaling the number of pixels that represent the length of 1 PDF canvas unit (usually 1/72 in). [1]_
+                This defines the resolution of the image. To convert a DPI value to a scale factor, multiply it by the size of 1 canvas unit in inches.
                 
                 .. [1] Since PDF 1.6, pages may define a so-called user unit. In this case, 1 canvas unit is equivalent to ``user_unit * (1/72)`` inches. pypdfium2 currently does not take this into account.
                 
@@ -430,30 +485,13 @@ class PdfPage (BitmapConvAliases):
             (ctypes array, str, (int, int)): Bitmap data, colour format, and image size.
             The colour format may be ``BGR``/``RGB``, ``BGRA``/``RGBA``, ``BGRX``/``RGBX``, or ``L``, depending on the parameters *colour*, *greyscale*, *rev_byteorder* and *prefer_bgrx*.
             Image size is given in pixels as a tuple of width and height.
-        
-        Tip:
-            To convert a DPI value to a scale factor, divide it by 72.
-        
-        Note:
-            Please take into account the following information regarding the *allocator* parameter:
-            
-            If you wish to render to an existing buffer that was not alloacted by ctypes itself, note that you may get a ctypes array representation of arbitrary memory using ``(ctypes.c_ubyte*n_bytes).from_address(mem_address)``, where *n_bytes* shall be the number of bytes to encompass, and *mem_address* the memory address of the first byte. This may be used to directly write the pixel data to a specific place in memory (e. g. a GUI widget buffer), avoiding unnecessary data copying.
-            
-            In this case, callers must ensure that ...
-            
-            * the buffer and its ctypes representation are large enough to hold the requested number of bytes
-            * the memory remains valid as long as the ctypes array is used
-            * the memory is freed once not needed anymore
         """
         
-        validate_colours(fill_colour, colour_scheme)
-        
         if force_bitmap_format in (None, pdfium.FPDFBitmap_Unknown):
-            cl_pdfium = auto_bitmap_format(fill_colour, greyscale, prefer_bgrx)
+            cl_pdfium = _auto_bitmap_format(fill_colour, greyscale, prefer_bgrx)
         else:
             cl_pdfium = force_bitmap_format
         
-        # attempting to use FPDF_REVERSE_BYTE_ORDER with FPDFBitmap_Gray is not only unnecessary, but also causes issues, so prohibit it (pdfium bug?)
         if cl_pdfium == pdfium.FPDFBitmap_Gray:
             rev_byteorder = False
         
@@ -528,7 +566,6 @@ class PdfPage (BitmapConvAliases):
         if colour_scheme is None:
             pdfium.FPDF_RenderPageBitmap(*render_args)
         else:
-            # rendering with colour scheme is only available as async variant at the moment
             
             ifsdk_pause = pdfium.IFSDK_PAUSE()
             ifsdk_pause.version = 1
@@ -541,12 +578,26 @@ class PdfPage (BitmapConvAliases):
             pdfium.FPDF_RenderPage_Close(self.raw)
         
         if draw_forms:
-            form_env = self.pdf.init_formenv()
-            pdfium.FPDF_FFLDraw(form_env, *render_args)
+            form_type = pdfium.FPDF_GetFormType(self.pdf.raw)  # consider moving to document ?
+            if form_type != pdfium.FORMTYPE_NONE:
+                form_env = self.pdf.init_formenv()
+                pdfium.FPDF_FFLDraw(form_env, *render_args)
         
         return buffer, cl_string, (width, height)
 
 
+def _auto_bitmap_format(fill_colour, greyscale, prefer_bgrx):
+    # no need to take alpha values of colour_scheme into account (drawings are additive)
+    if (fill_colour[3] < 255):
+        return pdfium.FPDFBitmap_BGRA
+    elif greyscale:
+        return pdfium.FPDFBitmap_Gray
+    elif prefer_bgrx:
+        return pdfium.FPDFBitmap_BGRx
+    else:
+        return pdfium.FPDFBitmap_BGR
+
+
 class ColourScheme:
     """
     Rendering colour scheme.
@@ -578,34 +629,3 @@ class ColourScheme:
         for key, value in self.colours.items():
             setattr(fpdf_cs, key, colour_tohex(value, rev_byteorder))
         return fpdf_cs
-
-
-class PdfPageObject:
-    """
-    Page object helper class.
-    
-    Attributes:
-        raw (FPDF_PAGEOBJECT): The underlying PDFium pageobject handle.
-        page (PdfPage): Reference to the page this pageobject belongs to.
-        level (int): Nesting level signifying the number of parent Form XObjects. Zero if the object is not nested in a Form XObject.
-        type (int): The type of the object (:data:`FPDF_PAGEOBJ_*`).
-    """
-    
-    def __init__(self, raw, page, level=0):
-        self.raw = raw
-        self.page = page
-        self.level = level
-        self.type = pdfium.FPDFPageObj_GetType(self.raw)
-    
-    def get_pos(self):
-        """
-        Get the position of the object on the page.
-        
-        Returns:
-            A tuple of four :class:`float` coordinates for left, bottom, right, and top.
-        """
-        left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
-        ret_code = pdfium.FPDFPageObj_GetBounds(self.raw, left, bottom, right, top)
-        if not ret_code:
-            raise PdfiumError("Locating the page object failed")
-        return (left.value, bottom.value, right.value, top.value)
